import os
import signal
import sys

import pytest

from anyio import create_task_group, fail_after, open_signal_receiver, run_sync_in_worker_thread

pytestmark = pytest.mark.anyio


@pytest.mark.skipif(sys.platform == 'win32', reason='Signal delivery cannot be tested on Windows')
async def test_receive_signals():
    async with open_signal_receiver(signal.SIGUSR1, signal.SIGUSR2) as sigiter:
        await run_sync_in_worker_thread(os.kill, os.getpid(), signal.SIGUSR1)
        await run_sync_in_worker_thread(os.kill, os.getpid(), signal.SIGUSR2)
        async with fail_after(1):
            assert await sigiter.__anext__() == signal.SIGUSR1
            assert await sigiter.__anext__() == signal.SIGUSR2


@pytest.mark.skipif(sys.platform == 'win32', reason='Signal delivery cannot be tested on Windows')
async def test_task_group_cancellation():
    async def signal_handler():
        async with open_signal_receiver(signal.SIGUSR1) as sigiter:
            async for v in sigiter:
                assert False

    async with create_task_group() as tg:
        await tg.spawn(signal_handler)
        await tg.cancel_scope.cancel()



from collections import deque
from dataclasses import InitVar, dataclass, field
from typing import Iterable

import pytest

from anyio import ClosedResourceError, EndOfStream
from anyio.abc import ByteReceiveStream, ByteSendStream, ObjectReceiveStream, ObjectSendStream
from anyio.streams.stapled import StapledByteStream, StapledObjectStream

pytestmark = pytest.mark.anyio


class TestStapledByteStream:
    @dataclass
    class DummyByteReceiveStream(ByteReceiveStream):
        data: InitVar[bytes]
        buffer: bytearray = field(init=False)
        _closed: bool = field(init=False, default=False)

        def __post_init__(self, data: bytes):
            self.buffer = bytearray(data)

        async def receive(self, max_bytes: int = 65536) -> bytes:
            if self._closed:
                raise ClosedResourceError

            data = bytes(self.buffer[:max_bytes])
            del self.buffer[:max_bytes]
            return data

        async def aclose(self) -> None:
            self._closed = True

    @dataclass
    class DummyByteSendStream(ByteSendStream):
        buffer: bytearray = field(init=False, default_factory=bytearray)
        _closed: bool = field(init=False, default=False)

        async def send(self, item: bytes) -> None:
            if self._closed:
                raise ClosedResourceError

            self.buffer.extend(item)

        async def aclose(self) -> None:
            self._closed = True

    @pytest.fixture
    def stapled(self):
        receive = self.DummyByteReceiveStream(b'hello, world')
        send = self.DummyByteSendStream()
        return StapledByteStream(send, receive)

    async def test_receive_send(self, stapled):
        assert await stapled.receive(3) == b'hel'
        assert await stapled.receive() == b'lo, world'
        assert await stapled.receive() == b''

        await stapled.send(b'how are you ')
        await stapled.send(b'today?')
        assert bytes(stapled.send_stream.buffer) == b'how are you today?'

    async def test_send_eof(self, stapled):
        await stapled.send_eof()
        await stapled.send_eof()
        with pytest.raises(ClosedResourceError):
            await stapled.send(b'world')

        assert await stapled.receive() == b'hello, world'

    async def test_aclose(self, stapled):
        await stapled.aclose()
        with pytest.raises(ClosedResourceError):
            await stapled.receive()
        with pytest.raises(ClosedResourceError):
            await stapled.send(b'')


class TestStapledObjectStream:
    @dataclass
    class DummyObjectReceiveStream(ObjectReceiveStream):
        data: InitVar[Iterable]
        buffer: deque = field(init=False)
        _closed: bool = field(init=False, default=False)

        def __post_init__(self, data: Iterable):
            self.buffer = deque(data)

        async def receive(self):
            if self._closed:
                raise ClosedResourceError
            if not self.buffer:
                raise EndOfStream

            return self.buffer.popleft()

        async def aclose(self) -> None:
            self._closed = True

    @dataclass
    class DummyObjectSendStream(ObjectSendStream):
        buffer: list = field(init=False, default_factory=list)
        _closed: bool = field(init=False, default=False)

        async def send(self, item) -> None:
            if self._closed:
                raise ClosedResourceError

            self.buffer.append(item)

        async def aclose(self) -> None:
            self._closed = True

    @pytest.fixture
    def stapled(self):
        receive = self.DummyObjectReceiveStream(['hello', 'world'])
        send = self.DummyObjectSendStream()
        return StapledObjectStream(send, receive)

    async def test_receive_send(self, stapled):
        assert await stapled.receive() == 'hello'
        assert await stapled.receive() == 'world'
        with pytest.raises(EndOfStream):
            await stapled.receive()

        await stapled.send('how are you ')
        await stapled.send('today?')
        assert stapled.send_stream.buffer == ['how are you ', 'today?']

    async def test_send_eof(self, stapled):
        await stapled.send_eof()
        await stapled.send_eof()
        with pytest.raises(ClosedResourceError):
            await stapled.send('world')

        assert await stapled.receive() == 'hello'
        assert await stapled.receive() == 'world'

    async def test_aclose(self, stapled):
        await stapled.aclose()
        with pytest.raises(ClosedResourceError):
            await stapled.receive()
        with pytest.raises(ClosedResourceError):
            await stapled.send(b'')



import threading
from typing import Any, Callable, Coroutine, Dict, Optional, TypeVar, cast

from ..abc import BlockingPortal, CapacityLimiter
from ._eventloop import get_asynclib, run, threadlocals

T_Retval = TypeVar('T_Retval')


async def run_sync_in_worker_thread(
        func: Callable[..., T_Retval], *args, cancellable: bool = False,
        limiter: Optional[CapacityLimiter] = None) -> T_Retval:
    """
    Call the given function with the given arguments in a worker thread.

    If the ``cancellable`` option is enabled and the task waiting for its completion is cancelled,
    the thread will still run its course but its return value (or any raised exception) will be
    ignored.

    :param func: a callable
    :param args: positional arguments for the callable
    :param cancellable: ``True`` to allow cancellation of the operation
    :param limiter: capacity limiter to use to limit the total amount of threads running
        (if omitted, the default limiter is used)
    :return: an awaitable that yields the return value of the function.

    """
    return await get_asynclib().run_sync_in_worker_thread(func, *args, cancellable=cancellable,
                                                          limiter=limiter)


def run_async_from_thread(func: Callable[..., Coroutine[Any, Any, T_Retval]], *args) -> T_Retval:
    """
    Call a coroutine function from a worker thread.

    :param func: a coroutine function
    :param args: positional arguments for the callable
    :return: the return value of the coroutine function

    """
    try:
        asynclib = threadlocals.current_async_module
    except AttributeError:
        raise RuntimeError('This function can only be run from an AnyIO worker thread')

    return asynclib.run_async_from_thread(func, *args)


def current_default_worker_thread_limiter() -> CapacityLimiter:
    """
    Return the capacity limiter that is used by default to limit the number of concurrent threads.

    :return: a capacity limiter object

    """
    return get_asynclib().current_default_thread_limiter()


def create_blocking_portal() -> BlockingPortal:
    """
    Create a portal for running functions in the event loop thread from external threads.

    Use this function in asynchronous code when you need to allow external threads access to the
    event loop where your asynchronous code is currently running.
    """
    return get_asynclib().BlockingPortal()


def start_blocking_portal(
        backend: str = 'asyncio',
        backend_options: Optional[Dict[str, Any]] = None) -> BlockingPortal:
    """
    Start a new event loop in a new thread and run a blocking portal in its main task.

    The parameters are the same as for :func:`~anyio.run`.

    :param backend: name of the backend
    :param backend_options: backend options
    :return: a blocking portal object

    """
    async def run_portal():
        nonlocal portal
        async with create_blocking_portal() as portal:
            event.set()
            await portal.sleep_until_stopped()

    portal: Optional[BlockingPortal]
    event = threading.Event()
    kwargs = {'func': run_portal, 'backend': backend, 'backend_options': backend_options}
    thread = threading.Thread(target=run, kwargs=kwargs)
    thread.start()
    event.wait()
    return cast(BlockingPortal, portal)
